"""Dataset class for defining data source, splitting, parsing, and iteration."""

import json
from functools import partial
from inspect import Parameter, signature
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Type, TypeVar, cast

import pandas as pd
from flytekit.core.tracker import TrackedInstance
from flytekit.extras.sqlite3.task import SQLite3Task
from sklearn.model_selection import train_test_split

from unionml.utils import inner_task

R = TypeVar("R")  # raw data
D = TypeVar("D")  # model-ready data


class Dataset(TrackedInstance):
    def __init__(
        self,
        name: str = "dataset",
        *,
        features: Optional[List[str]] = None,
        targets: Optional[List[str]] = None,
        test_size: float = 0.2,
        shuffle: bool = True,
        random_state: int = 12345,
    ):
        super().__init__()
        self.name = name
        self._features = [] if features is None else features
        self._targets = targets
        self._test_size = test_size
        self._shuffle = shuffle
        self._random_state = random_state

        # default component functions
        self._loader = self._default_loader
        self._splitter = self._default_splitter
        self._parser = self._default_parser
        self._parser_feature_key: int = 0  # assume that first element of parser tuple output contains features
        self._feature_loader = self._default_feature_loader
        self._feature_transformer = self._default_feature_transformer

        self._reader = None
        self._reader_input_types: Optional[List[Parameter]] = None
        self._reader_return_type: Optional[Dict[str, Type]] = None
        self._dataset_task = None

    def reader(self, fn=None, **reader_task_kwargs):
        """Register a reader function for getting data from some external source."""
        if fn is None:
            return partial(self.reader, **reader_task_kwargs)
        self._reader = fn
        self._reader_task_kwargs = reader_task_kwargs
        return fn

    def loader(self, fn):
        """Register an optional loader function for loading data into memory for model training.

        This function should take the output of the reader function and return the data structure needed
        for model training.
        """
        self._loader = fn
        return fn

    def splitter(self, fn):
        """Register an optional splitter function that partitions data into training and test sets."""
        self._splitter = fn
        return fn

    def parser(self, fn, feature_key: int = 0):
        """Register an optional parser function that produces a tuple of features and targets."""
        self._parser = fn
        self._parser_feature_key = feature_key
        return fn

    def feature_loader(self, fn):
        """Register an optional feature loader that loads data from some serialized format into raw features.

        This function handles prediction cases in two contexts:
        1. When the `unionml predict` cli command is invoked with the --features flag.
        2. When the FastAPI app `/predict/` endpoint is invoked with features passed in as JSON or string encoding.

        And it should return the data structure needed for model training.
        """
        self._feature_loader = fn
        return fn

    def feature_transformer(self, fn):
        """Register an optional feature transformer that performs pre-processing on features before prediction.

        This function handles prediction cases in three contexts:
        1. When the `unionml predict` cli command is invoked with the --features flag.
        2. When the FastAPI app `/predict/` endpoint is invoked with features passed in as JSON or string encoding.
        3. When the `model.predict` or `model.remote_predict` functions are invoked.
        ."""
        self._feature_transformer = fn
        return fn

    @property
    def splitter_kwargs(self):
        return {
            "test_size": self._test_size,
            "shuffle": self._shuffle,
            "random_state": self._random_state,
        }

    @property
    def parser_kwargs(self):
        return {
            "features": self._features,
            "targets": self._targets,
        }

    def dataset_task(self):
        if self._dataset_task:
            return self._dataset_task

        reader_sig = signature(self._reader)

        # TODO: make sure return type is not None
        @inner_task(
            unionml_obj=self,
            input_parameters=reader_sig.parameters,
            return_annotation=NamedTuple("ReaderOutput", data=reader_sig.return_annotation),
            **self._reader_task_kwargs,
        )
        def dataset_task(*args, **kwargs):
            return self._reader(*args, **kwargs)

        self._dataset_task = dataset_task
        return dataset_task

    @property
    def literal_data_workflow_name(self):
        return f"{self.name}.literal_data"

    @property
    def literal_features_workflow_name(self):
        return f"{self.name}.literal_features"

    @property
    def data_workflow_name(self):
        return f"{self.name}.data"

    @property
    def features_workflow_name(self):
        return f"{self.name}.features"

    def get_data(self, raw_data):
        data = self._loader(raw_data)
        splits = self._splitter(data, **self.splitter_kwargs)
        if len(splits) == 1:
            return {"train": self._parser(splits[0], **self.parser_kwargs)}

        train_split, test_split = splits
        # TODO: make this more generic so as to include a validation split
        train_data = self._parser(train_split, **self.parser_kwargs)
        test_data = self._parser(test_split, **self.parser_kwargs)
        return {
            "train": train_data,
            "test": test_data,
        }

    def get_features(self, features):
        parsed_data = self._parser(self._feature_loader(features), self._features, self._targets)
        features = parsed_data[self._parser_feature_key]
        return self._feature_transformer(features)

    @property
    def reader_input_types(self) -> Optional[List[Parameter]]:
        if self._reader and self._reader_input_types is None:
            return [p for p in signature(self._reader).parameters]
        return self._reader_input_types

    @property
    def reader_return_type(self) -> Dict[str, Type]:
        if self._reader and self._reader_return_type is None:
            return {"data": signature(self._reader).return_annotation}
        elif self._reader_return_type is not None:
            return self._reader_return_type
        raise ValueError(
            "reader_return_type is not defined. Please define a @dataset.reader function with an output annotation."
        )

    @classmethod
    def _from_flytekit_task(
        cls,
        task,
        *args,
        **kwargs,
    ) -> "Dataset":
        dataset = cls(*args, **kwargs)
        dataset._dataset_task = task
        dataset._reader_return_type = task.python_interface.outputs
        dataset._reader_input_types = [
            Parameter(k, Parameter.KEYWORD_ONLY, annotation=v) for k, v in task.python_interface.inputs.items()
        ]
        return dataset

    @classmethod
    def from_sqlite_task(
        cls,
        task: SQLite3Task,
        *args,
        **kwargs,
    ) -> "Dataset":
        return cls._from_flytekit_task(task, *args, **kwargs)

    @classmethod
    def from_sqlalchemy_task(
        cls,
        task: "flytekitplugins.sqlalchemy.SQLAlchemyTask",  # type: ignore
        *args,
        **kwargs,
    ) -> "Dataset":
        return cls._from_flytekit_task(task, *args, **kwargs)

    def _default_loader(self, data: R) -> R:
        [(_, data_type)] = self.reader_return_type.items()
        if data_type is pd.DataFrame:
            return pd.DataFrame(data)
        return data

    def _default_splitter(
        self,
        data: D,
        test_size: float,
        shuffle: bool,
        random_state: int,
    ) -> Tuple[D, ...]:
        if not isinstance(data, pd.DataFrame):
            return (data,)
        return train_test_split(data, test_size=test_size, random_state=random_state, shuffle=shuffle)

    def _default_parser(self, data: D, features: Optional[List[str]], targets: Optional[List[str]]) -> Tuple[D, ...]:
        if not isinstance(data, pd.DataFrame):
            return (data,)

        if features is not None and targets is not None:
            features = [col for col in data if col not in targets]
        try:
            target_data = data[targets]
        except KeyError:
            target_data = pd.DataFrame()
        return data[features], target_data

    def _default_feature_loader(self, features: Any) -> R:
        if isinstance(features, Path):
            with features.open() as f:
                features = json.load(f)

        [(_, data_type)] = self.reader_return_type.items()
        if data_type is pd.DataFrame:
            return pd.DataFrame(features)
        return features

    def _default_feature_transformer(self, features: R) -> D:
        """
        By default this is just a pass-through function. The user can choose to override it with
        the `@dataset.feature_transformer` decorator.
        """
        return cast(D, features)



import importlib
from typing import List

from flytekit.core.base_task import TaskResolverMixin
from flytekit.core.context_manager import SerializationSettings
from flytekit.core.python_auto_container import PythonAutoContainerTask
from flytekit.core.tracker import TrackedInstance


class TaskResolver(TrackedInstance, TaskResolverMixin):
    """Task Resolver for unionml"""

    def name(self) -> str:
        return "TaskResolver"

    def load_task(self, loader_args: List[str]) -> PythonAutoContainerTask:
        _, app_module, _, unionml_obj, _, task_name, *_ = loader_args

        _unionml_obj = getattr(importlib.import_module(app_module), unionml_obj)
        task_method = getattr(_unionml_obj, task_name)
        return task_method()

    def loader_args(self, settings: SerializationSettings, task: PythonAutoContainerTask) -> List[str]:
        return [
            "app-module",
            task.task_function.__unionml_object__.instantiated_in,
            "unionml-obj-name",
            task.task_function.__unionml_object__.lhs,
            "task-name",
            task.task_function.__name__,
        ]


task_resolver = TaskResolver()



"""Model class for defining training, evaluation, and prediction."""

import inspect
import os
from collections import OrderedDict
from dataclasses import asdict, dataclass, field, is_dataclass, make_dataclass
from functools import partial
from inspect import Parameter, signature
from typing import IO, Any, Callable, Dict, List, NamedTuple, Optional, Tuple, Type, Union

import joblib
import sklearn
from dataclasses_json import dataclass_json
from fastapi import FastAPI
from flytekit import Workflow
from flytekit.configuration import Config
from flytekit.core.tracker import TrackedInstance
from flytekit.remote import FlyteRemote
from flytekit.remote.executions import FlyteWorkflowExecution

from unionml.dataset import Dataset
from unionml.utils import inner_task, is_keras_model, is_pytorch_model


@dataclass
class BaseHyperparameters:
    """Hyperparameter base class"""

    pass


class ModelArtifact(NamedTuple):
    """Model artifact, containing a specific model object and optional metrics associated with it."""

    model_object: Any
    hyperparameters: Optional[Union[BaseHyperparameters, dict]] = None
    metrics: Optional[Dict[str, float]] = None


class Model(TrackedInstance):
    def __init__(
        self,
        name: str = "model",
        init: Union[Type, Callable] = None,
        *,
        dataset: Dataset,
        hyperparameter_config: Optional[Dict[str, Type]] = None,
    ):
        super().__init__()
        self.name = name
        self._init_callable = init
        self._hyperparameter_config = hyperparameter_config
        self._dataset = dataset
        self._artifact: Optional[ModelArtifact] = None

        # default component functions
        self._init = self._default_init
        self._saver = self._default_saver
        self._loader = self._default_loader

        # properties needed for deployment
        self._image_name: Optional[str] = None
        self._config_file: Optional[str] = None
        self._registry: Optional[str] = None
        self._dockerfile: Optional[str] = None
        self._project: Optional[str] = None
        self._domain: Optional[str] = None
        self.__remote__: Optional[FlyteRemote] = None

        if self._dataset.name is None:
            self._dataset.name = f"{self.name}.dataset"

        # unionml-compiled tasks
        self._train_task = None
        self._predict_task = None
        self._predict_from_features_task = None

        # user-provided task kwargs
        self._train_task_kwargs = None
        self._predict_task_kwargs = None

        # dynamically defined types
        self._hyperparameter_type: Optional[Type] = None

    @property
    def artifact(self) -> Optional[ModelArtifact]:
        return self._artifact

    @artifact.setter
    def artifact(self, new_value: ModelArtifact):
        self._artifact = new_value

    @property
    def hyperparameter_type(self) -> Type:
        if self._hyperparameter_type is not None:
            return self._hyperparameter_type

        hyperparameter_fields: List[Any] = []
        if self._hyperparameter_config is None:
            # extract types from the init callable that instantiates a new model
            model_obj_sig = signature(self._init_callable)  # type: ignore

            # if any of the arguments are not type-annotated, default to using an untyped dictionary
            if any(p.annotation is inspect._empty for p in model_obj_sig.parameters.values()):
                return dict

            for hparam_name, hparam in model_obj_sig.parameters.items():
                hyperparameter_fields.append((hparam_name, hparam.annotation, field(default=hparam.default)))
        else:
            # extract types from hyperparameters Model init argument
            for hparam_name, hparam_type in self._hyperparameter_config.items():
                hyperparameter_fields.append((hparam_name, hparam_type))

        self._hyperparameter_type = dataclass_json(
            make_dataclass("Hyperparameters", hyperparameter_fields, bases=(BaseHyperparameters,))
        )
        return self._hyperparameter_type

    @property
    def config_file(self) -> Optional[str]:
        return self._config_file

    @property
    def registry(self) -> Optional[str]:
        return self._registry

    @property
    def dockerfile(self) -> Optional[str]:
        return self._dockerfile

    @property
    def train_workflow_name(self):
        return f"{self.name}.train"

    @property
    def predict_workflow_name(self):
        return f"{self.name}.predict"

    @property
    def predict_from_features_workflow_name(self):
        return f"{self.name}.predict_from_features"

    def init(self, fn):
        self._init = fn
        return self._init

    def trainer(self, fn=None, **train_task_kwargs):
        if fn is None:
            return partial(self.trainer, **train_task_kwargs)
        self._trainer = fn
        self._train_task_kwargs = train_task_kwargs
        return self._trainer

    def predictor(self, fn=None, **predict_task_kwargs):
        if fn is None:
            return partial(self.predictor, **predict_task_kwargs)
        self._predictor = fn
        self._predict_task_kwargs = predict_task_kwargs
        return self._predictor

    def evaluator(self, fn):
        self._evaluator = fn
        return self._evaluator

    def saver(self, fn):
        self._saver = fn
        return self._saver

    def loader(self, fn):
        self._loader = fn
        return self._loader

    @property
    def trainer_params(self):
        return {
            name: param
            for name, param in signature(self._trainer).parameters.items()
            if param.kind == Parameter.KEYWORD_ONLY
        }

    def train_workflow(self):
        dataset_task = self._dataset.dataset_task()
        train_task = self.train_task()

        [
            hyperparam_arg,
            hyperparam_type,
        ], *_ = train_task.python_interface.inputs.items()

        wf = Workflow(name=self.train_workflow_name)

        # add hyperparameter argument
        wf.add_workflow_input(hyperparam_arg, hyperparam_type)

        # add dataset.reader arguments
        for arg, type in dataset_task.python_interface.inputs.items():
            wf.add_workflow_input(arg, type)

        # add training keyword-only arguments
        trainer_param_types = {k: v.annotation for k, v in self.trainer_params.items()}
        for arg, type in trainer_param_types.items():
            wf.add_workflow_input(arg, type)

        dataset_node = wf.add_entity(
            dataset_task,
            **{k: wf.inputs[k] for k in dataset_task.python_interface.inputs},
        )
        train_node = wf.add_entity(
            train_task,
            **{
                hyperparam_arg: wf.inputs[hyperparam_arg],
                **dataset_node.outputs,
                **{arg: wf.inputs[arg] for arg in trainer_param_types},
            },
        )
        wf.add_workflow_output("model_object", train_node.outputs["model_object"])
        wf.add_workflow_output("hyperparameters", train_node.outputs["hyperparameters"])
        wf.add_workflow_output("metrics", train_node.outputs["metrics"])
        return wf

    def predict_workflow(self):
        dataset_task = self._dataset.dataset_task()
        predict_task = self.predict_task()

        wf = Workflow(name=self.predict_workflow_name)
        model_arg_name, *_ = predict_task.python_interface.inputs.keys()
        wf.add_workflow_input("model_object", predict_task.python_interface.inputs[model_arg_name])
        for arg, type in dataset_task.python_interface.inputs.items():
            wf.add_workflow_input(arg, type)

        dataset_node = wf.add_entity(
            dataset_task,
            **{k: wf.inputs[k] for k in dataset_task.python_interface.inputs},
        )
        predict_node = wf.add_entity(
            predict_task, **{"model_object": wf.inputs["model_object"], **dataset_node.outputs}
        )
        for output_name, promise in predict_node.outputs.items():
            wf.add_workflow_output(output_name, promise)
        return wf

    def predict_from_features_workflow(self):
        predict_task = self.predict_from_features_task()

        wf = Workflow(name=self.predict_from_features_workflow_name)
        for i, (arg, type) in enumerate(predict_task.python_interface.inputs.items()):
            # assume that the first argument is the model object
            wf.add_workflow_input("model_object" if i == 0 else arg, type)

        predict_node = wf.add_entity(predict_task, **{k: wf.inputs[k] for k in wf.inputs})
        for output_name, promise in predict_node.outputs.items():
            wf.add_workflow_output(output_name, promise)
        return wf

    def train_task(self):
        if self._train_task:
            return self._train_task

        # make sure hyperparameter type signature is correct
        *_, hyperparameters_param = signature(self._init).parameters.values()
        hyperparameters_param = hyperparameters_param.replace(annotation=self.hyperparameter_type)

        # assume that reader_return_type is a dict with only a single entry
        [(data_arg_name, data_arg_type)] = self._dataset.reader_return_type.items()

        # get keyword-only training args
        @inner_task(
            unionml_obj=self,
            input_parameters=OrderedDict(
                [
                    (p.name, p)
                    for p in [
                        hyperparameters_param,
                        Parameter(
                            data_arg_name,
                            kind=Parameter.KEYWORD_ONLY,
                            annotation=data_arg_type,
                        ),
                        *self.trainer_params.values(),
                    ]
                ]
            ),
            return_annotation=NamedTuple(
                "ModelArtifact",
                model_object=signature(self._trainer).return_annotation,
                hyperparameters=self.hyperparameter_type,
                metrics=Dict[str, signature(self._evaluator).return_annotation],
            ),
            **({} if self._train_task_kwargs is None else self._train_task_kwargs),
        )
        def train_task(**kwargs):
            hyperparameters = kwargs["hyperparameters"]
            raw_data = kwargs[data_arg_name]
            trainer_kwargs = {p: kwargs[p] for p in self.trainer_params}

            hyperparameters_dict = asdict(hyperparameters) if is_dataclass(hyperparameters) else hyperparameters
            training_data = self._dataset.get_data(raw_data)
            model_object = self._trainer(
                self._init(hyperparameters=hyperparameters_dict),
                *training_data["train"],
                **trainer_kwargs,
            )
            metrics = {
                split_key: self._evaluator(model_object, *training_data[split_key]) for split_key in training_data
            }
            return model_object, hyperparameters, metrics

        self._train_task = train_task
        return train_task

    def predict_task(self):
        if self._predict_task:
            return self._predict_task

        predictor_sig = signature(self._predictor)
        model_param, *_ = predictor_sig.parameters.values()
        model_param = model_param.replace(name="model_object")

        # assume that reader_return_type is a dict with only a single entry
        [(data_arg_name, data_arg_type)] = self._dataset.reader_return_type.items()
        data_param = Parameter(data_arg_name, kind=Parameter.KEYWORD_ONLY, annotation=data_arg_type)

        # TODO: make sure return type is not None
        @inner_task(
            unionml_obj=self,
            input_parameters=OrderedDict([(p.name, p) for p in [model_param, data_param]]),
            return_annotation=predictor_sig.return_annotation,
            **self._predict_task_kwargs,
        )
        def predict_task(model_object, **kwargs):
            parsed_data = self._dataset._parser(kwargs[data_arg_name], **self._dataset.parser_kwargs)
            features = parsed_data[self._dataset._parser_feature_key]
            return self._predictor(model_object, features)

        self._predict_task = predict_task
        return predict_task

    def predict_from_features_task(self):
        if self._predict_from_features_task:
            return self._predict_from_features_task

        predictor_sig = signature(self._predictor)
        model_param, *_ = predictor_sig.parameters.values()
        model_param = model_param.replace(name="model_object")

        # assume that reader_return_type is a dict with only a single entry
        [(_, data_arg_type)] = self._dataset.reader_return_type.items()
        data_param = Parameter("features", kind=Parameter.KEYWORD_ONLY, annotation=data_arg_type)

        @inner_task(
            unionml_obj=self,
            input_parameters=OrderedDict([("model_object", model_param), ("features", data_param)]),
            return_annotation=predictor_sig.return_annotation,
            **self._predict_task_kwargs,
        )
        def predict_from_features_task(model_object, features):
            return self._predictor(model_object, self._dataset.get_features(features))

        self._predict_from_features_task = predict_from_features_task
        return predict_from_features_task

    def train(
        self,
        hyperparameters: Optional[Dict[str, Any]] = None,
        trainer_kwargs: Optional[Dict[str, Any]] = None,
        **reader_kwargs,
    ) -> Tuple[Any, Any]:
        trainer_kwargs = {} if trainer_kwargs is None else trainer_kwargs
        model_obj, hyperparameters, metrics = self.train_workflow()(
            hyperparameters=self.hyperparameter_type(**({} if hyperparameters is None else hyperparameters)),
            **{**reader_kwargs, **trainer_kwargs},
        )
        self.artifact = ModelArtifact(model_obj, hyperparameters, metrics)
        return model_obj, metrics

    def predict(
        self,
        features: Any = None,
        **reader_kwargs,
    ):
        if self.artifact is None:
            raise RuntimeError(
                "ModelArtifact not found. You must train a model first with the `train` method before generating "
                "predictions."
            )
        if features is None:
            return self.predict_workflow()(model_object=self.artifact.model_object, **reader_kwargs)
        return self.predict_from_features_workflow()(model_object=self.artifact.model_object, features=features)

    def save(self, file, *args, **kwargs):
        if self.artifact is None:
            raise AttributeError("`artifact` property is None. Call the `train` method to train a model first")
        return self._saver(self.artifact.model_object, self.artifact.hyperparameters, file, *args, **kwargs)

    def load(self, file, *args, **kwargs):
        return self._loader(file, *args, **kwargs)

    def serve(self, app: FastAPI, remote: bool = False, model_version: str = "latest"):
        """Create a FastAPI serving app."""
        from unionml.fastapi import serving_app

        serving_app(self, app, remote=remote, model_version=model_version)

    def remote(
        self,
        registry: Optional[str] = None,
        image_name: str = None,
        dockerfile: str = "Dockerfile",
        config_file: Optional[str] = None,
        project: Optional[str] = None,
        domain: Optional[str] = None,
    ):
        self._config_file = config_file
        self._registry = registry
        self._image_name = image_name
        self._dockerfile = dockerfile
        self._project = project
        self._domain = domain

    @property
    def _remote(self) -> Optional[FlyteRemote]:
        if self.__remote__ is not None:
            return self.__remote__

        config = Config.auto(config_file=self._config_file)
        if config.platform.endpoint.startswith("localhost"):
            config = Config.for_sandbox()

        self.__remote__ = FlyteRemote(
            config=config,
            default_project=self._project,
            default_domain=self._domain,
        )
        return self.__remote__

    def remote_deploy(self):
        """Deploy model services to a Flyte backend."""
        from unionml import remote

        app_version = remote.get_app_version()
        image = remote.get_image_fqn(self, app_version, self._image_name)

        os.environ["FLYTE_INTERNAL_IMAGE"] = image or ""
        _remote = self._remote

        remote.create_project(_remote, self._project)
        if _remote.config.platform.endpoint.startswith("localhost"):
            # assume that a localhost flyte_admin_url means that we want to use Flyte sandbox
            remote.sandbox_docker_build(self, image)
        else:
            remote.docker_build_push(self, image)

        args = [_remote._default_project, _remote._default_domain, app_version]
        for wf in [
            self.train_workflow(),
            self.predict_workflow(),
            self.predict_from_features_workflow(),
        ]:
            remote.deploy_wf(wf, _remote, image, *args)

    def remote_train(
        self,
        app_version: str = None,
        wait: bool = True,
        *,
        hyperparameters: Optional[Dict[str, Any]] = None,
        trainer_kwargs: Optional[Dict[str, Any]] = None,
        **reader_kwargs,
    ) -> Union[ModelArtifact, FlyteWorkflowExecution]:
        if self._remote is None:
            raise RuntimeError("First configure the remote client with the `Model.remote` method")

        from unionml import remote

        app_version = app_version or remote.get_app_version()
        train_wf = self._remote.fetch_workflow(name=self.train_workflow_name, version=app_version)
        execution = self._remote.execute(
            train_wf,
            inputs={
                "hyperparameters": self.hyperparameter_type(**({} if hyperparameters is None else hyperparameters)),
                **{**reader_kwargs, **({} if trainer_kwargs is None else trainer_kwargs)},  # type: ignore
            },
            project=self._remote.default_project,
            domain=self._remote.default_domain,
            wait=wait,
            type_hints={"hyperparameters": self.hyperparameter_type},
        )
        console_url = self._remote.generate_console_url(execution)
        print(
            f"Executing {train_wf.id.name}, execution name: {execution.id.name}."
            f"\nGo to {console_url} to see the execution in the console."
        )
        if not wait:
            return execution

        self.remote_load(execution)
        return self.artifact

    def remote_predict(
        self,
        app_version: str = None,
        model_version: str = None,
        wait: bool = True,
        *,
        features: Any = None,
        **reader_kwargs,
    ) -> Union[Any, FlyteWorkflowExecution]:
        if self._remote is None:
            raise RuntimeError("First configure the remote client with the `Model.remote` method")

        from unionml import remote

        app_version = app_version or remote.get_app_version()
        model_artifact = remote.get_model_artifact(self, app_version, model_version)

        if (features is not None and len(reader_kwargs) > 0) or (features is None and len(reader_kwargs) == 0):
            raise ValueError("You must provide only one of `features` or `reader_kwargs`")

        inputs = {"model_object": model_artifact.model_object}
        if features is None:
            workflow_name = self.predict_workflow_name
            inputs.update(reader_kwargs)
            type_hints = {}
        else:
            workflow_name = self.predict_from_features_workflow_name
            inputs.update({"features": features})
            type_hints = {"features": [*self._dataset.reader_return_type.values()][0]}

        predict_wf = self._remote.fetch_workflow(
            self._remote._default_project,
            self._remote._default_domain,
            workflow_name,
            app_version,
        )
        execution = self._remote.execute(
            predict_wf,
            inputs=inputs,
            project=self._remote.default_project,
            domain=self._remote.default_domain,
            wait=wait,
            type_hints=type_hints,
        )
        console_url = self._remote.generate_console_url(execution)
        print(
            f"Executing {predict_wf.id.name}, execution name: {execution.id.name}."
            f"\nGo to {console_url} to see the execution in the console."
        )
        if not wait:
            return execution
        predictions, *_ = execution.outputs.values()
        return predictions

    def remote_wait(self, execution: FlyteWorkflowExecution, **kwargs) -> Any:
        if self._remote is None:
            raise ValueError("You must call `model.remote` to attach a remote backend to this model.")
        return self._remote.wait(execution, **kwargs)

    def remote_load(self, execution: FlyteWorkflowExecution):
        if self._remote is None:
            raise ValueError("You must call `model.remote` to attach a remote backend to this model.")
        if not execution.is_done:
            print(f"Waiting for execution {execution.id.name} to complete...")
            execution = self.remote_wait(execution)
            print("Done.")

        with self._remote.remote_context():
            self.artifact = ModelArtifact(
                execution.outputs["model_object"],
                execution.outputs["hyperparameters"],
                execution.outputs["metrics"],
            )

    def remote_list_model_versions(self, app_version: str = None, limit: int = 10) -> List[str]:
        from unionml import remote

        app_version = app_version or remote.get_app_version()
        return remote.list_model_versions(self, app_version=app_version, limit=limit)

    def remote_fetch_predictions(self, execution: FlyteWorkflowExecution) -> Any:
        if self._remote is None:
            raise ValueError("You must call `model.remote` to attach a remote backend to this model.")
        execution = self._remote.wait(execution)
        predictions, *_ = execution.outputs.values()
        return predictions

    def _default_init(self, hyperparameters: dict) -> Any:
        if self._init_callable is None:
            raise ValueError(
                "When using the _default_init method, you must specify the init argument to the Model constructor."
            )
        return self._init_callable(**hyperparameters)

    def _default_saver(
        self,
        model_obj: Any,
        hyperparameters: Union[dict, BaseHyperparameters],
        file: Union[str, os.PathLike, IO],
        *args,
        **kwargs,
    ) -> Any:
        init = self._init_callable if self._init == self._default_init else self._init or self._init_callable
        model_type = init if inspect.isclass(init) else signature(init).return_annotation if init is not None else init
        hyperparameters = asdict(hyperparameters) if is_dataclass(hyperparameters) else hyperparameters
        if isinstance(model_obj, sklearn.base.BaseEstimator):
            return joblib.dump({"model_obj": model_obj, "hyperparameters": hyperparameters}, file, *args, **kwargs)
        elif is_pytorch_model(model_type):
            import torch

            torch.save(
                {"model_obj": model_obj.state_dict(), "hyperparameters": hyperparameters},
                file,
                *args,
                **kwargs,
            )
            return file
        elif is_keras_model(model_type):
            model_obj.save(file, *args, **kwargs)
            return file

        raise NotImplementedError(
            f"Default saver not defined for type {type(model_obj)}. Use the Model.saver decorator to define one."
        )

    def _default_loader(self, file: Union[str, os.PathLike, IO], *args, **kwargs) -> Any:
        init = self._init_callable if self._init == self._default_init else self._init or self._init_callable
        model_type = init if inspect.isclass(init) else signature(init).return_annotation if init is not None else init

        if issubclass(model_type, sklearn.base.BaseEstimator):
            deserialized_model = joblib.load(file, *args, **kwargs)
            return deserialized_model["model_obj"]
        elif is_pytorch_model(model_type):
            import torch

            deserialized_model = torch.load(file, *args, **kwargs)
            model = model_type(**deserialized_model["hyperparameters"])
            model.load_state_dict(deserialized_model["model_obj"])
            return model
        elif is_keras_model(model_type):
            from tensorflow import keras

            return keras.models.load_model(file)

        raise NotImplementedError(
            f"Default loader not defined for type {model_type}. Use the Model.loader decorator to define one."
        )



import logging
import re

from unionml.dataset import Dataset
from unionml.model import Model


# silence FlyteRemote beta warning since unionml won't expose it to end users.
class FlyteRemoteFilter(logging.Filter):
    def filter(self, record):
        return not re.match("^This feature is still in beta.+", record.getMessage())


# silence logger warnings having to do with PickleFile, since unionml allows this
# by default.
class PickleFilter(logging.Filter):
    def filter(self, record):
        return not re.match(".+Flyte will default to use PickleFile as the transport.+", record.getMessage())


flytekit_logger = logging.getLogger("flytekit")
flytekit_logger.addFilter(PickleFilter())

flytekit_remote_logger = flytekit_logger.getChild("remote")
flytekit_remote_logger.addFilter(FlyteRemoteFilter())


__version__ = "0.1.0b0"



"""Module for flyte remote helper functions."""

import importlib
import logging
import typing

import docker
import git
from flytekit.configuration import ImageConfig, SerializationSettings
from flytekit.models import filters
from flytekit.models.admin.common import Sort
from flytekit.models.project import Project
from flytekit.remote import FlyteRemote, FlyteWorkflowExecution

from unionml.model import Model, ModelArtifact

IMAGE_NAME = "unionml"
FLYTE_SANDBOX_CONTAINER_NAME = "flyte-sandbox"


logger = logging.getLogger("unionml")
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s"))
logger.addHandler(handler)


def get_model(app: str, reload: bool = False) -> Model:
    module_name, model_var = app.split(":")
    module = importlib.import_module(module_name)
    if reload:
        importlib.reload(module)
    return getattr(module, model_var)


def create_project(remote: FlyteRemote, project: typing.Optional[str]):
    project = project or remote.default_project
    projects, _ = remote.client.list_projects_paginated(filters=[filters.Equal("project.name", project)])
    if not projects:
        remote.client.register_project(Project(id=project, name=project, description=project))


def get_app_version() -> str:
    repo = git.Repo(".", search_parent_directories=True)
    commit = repo.rev_parse("HEAD")
    return commit.hexsha


def get_image_fqn(model: Model, app_version: str, image_name: typing.Optional[str] = None) -> str:
    image_name = IMAGE_NAME if image_name is None else image_name
    if model.registry is None:
        image_uri = image_name
    else:
        image_uri = f"{model.registry}/{image_name}"
    return f"{image_uri}:{model.name.replace('_', '-')}-{app_version}"


def sandbox_docker_build(model: Model, image_fqn: str):
    logger.info("Building docker container in flyte demo cluster.")
    client = docker.from_env()
    sandbox_container = None
    for container in client.containers.list():
        if container.name == FLYTE_SANDBOX_CONTAINER_NAME:
            sandbox_container = container

    if sandbox_container is None:
        raise RuntimeError(
            "Cannot find Flyte Demo Cluster. Make sure to install flytectl and create a sandbox with "
            "`flytectl demo start --source .`"
        )

    _, build_logs = sandbox_container.exec_run(["docker", "build", "/root", "--tag", image_fqn], stream=True)
    for line in build_logs:
        logger.info(line.decode().strip())


def docker_build_push(model: Model, image_fqn: str) -> docker.models.images.Image:
    if model.registry is None:
        raise ValueError("You must specify a registry in `model.remote` when deploying to a remote cluster.")

    client = docker.from_env()

    logger.info(f"Building image: {image_fqn}")
    build_logs = client.api.build(
        path=".",
        dockerfile=model.dockerfile,
        tag=image_fqn,
        rm=True,
    )
    for line in build_logs:
        logger.info(line.decode().strip())

    for line in client.api.push(image_fqn, stream=True, decode=True):
        logger.info(line)


def deploy_wf(wf, remote: FlyteRemote, image: str, project: str, domain: str, version: str):
    """Register all tasks, workflows, and launchplans needed to execute the workflow."""
    logger.info(f"Deploying workflow {wf.name}")
    serialization_settings = SerializationSettings(
        project=project,
        domain=domain,
        image_config=ImageConfig.auto(img_name=image),
    )
    remote.register_workflow(wf, serialization_settings, version)


def get_model_execution(
    model: Model,
    app_version: typing.Optional[str] = None,
    model_version: typing.Optional[str] = "latest",
) -> FlyteWorkflowExecution:
    if model._remote is None:
        raise RuntimeError("You need to configure the remote client with the `Model.remote` method")

    app_version = app_version or get_app_version()
    train_wf = model._remote.fetch_workflow(
        model._remote._default_project,
        model._remote._default_domain,
        model.train_workflow_name,
        app_version,
    )
    if model_version is not None and model_version != "latest":
        execution = model._remote.fetch_execution(
            project=train_wf.id.project,
            domain=train_wf.id.domain,
            name=model_version,
        )
    else:
        [execution, *_], _ = model._remote.client.list_executions_paginated(
            train_wf.id.project,
            train_wf.id.domain,
            limit=1,
            filters=[
                filters.Equal("launch_plan.name", train_wf.id.name),
                filters.Equal("phase", "SUCCEEDED"),
            ],
            sort_by=Sort("created_at", Sort.Direction.DESCENDING),
        )
        execution = FlyteWorkflowExecution.promote_from_model(execution)
    model._remote.sync(execution)
    return execution


def get_model_artifact(
    model: Model,
    app_version: typing.Optional[str] = None,
    model_version: typing.Optional[str] = "latest",
) -> ModelArtifact:
    execution = get_model_execution(model, app_version, model_version)
    model.remote_load(execution)
    assert model.artifact is not None
    return model.artifact


def list_model_versions(model: Model, app_version: typing.Optional[str] = None, limit: int = 10) -> typing.List[str]:
    if model._remote is None:
        raise RuntimeError("You need to configure the remote client with the `Model.remote` method")

    app_version = app_version or get_app_version()
    train_wf = model._remote.fetch_workflow(
        model._remote._default_project,
        model._remote._default_domain,
        model.train_workflow_name,
        app_version,
    )
    executions, _ = model._remote.client.list_executions_paginated(
        train_wf.id.project,
        train_wf.id.domain,
        limit=limit,
        filters=[
            filters.Equal("launch_plan.name", train_wf.id.name),
            filters.Equal("phase", "SUCCEEDED"),
        ],
        sort_by=Sort("created_at", Sort.Direction.DESCENDING),
    )
    return [x.id.name for x in executions]



import typing
from functools import partial, wraps
from inspect import Parameter, signature

from flytekit import task

from unionml.task_resolver import task_resolver


def inner_task(
    fn=None,
    *,
    unionml_obj,
    input_parameters=None,
    return_annotation=None,
    **task_kwargs,
):
    """A flytekit task defined within a Dataset or Model class.

    This wrapper does the following:
    - makes sure the wrapper function:
      - has the same signature as the original function
      - OR it takes on the signature specified by the ``input_parameters`` and ``return_annotation`` arguments
        if they are provided
    - renames the wrapped function to ``task_name``.
    - assigns an ``unionml_obj`` to the function object.
    - converts the wrapper function into a flytekit task, using the unionml task resolver.
    """

    if fn is None:
        return partial(
            inner_task,
            unionml_obj=unionml_obj,
            input_parameters=input_parameters,
            return_annotation=return_annotation,
            **task_kwargs,
        )

    @wraps(fn)
    def wrapper(*args, **kwargs):
        return fn(*args, **kwargs)

    fn_sig = signature(fn)

    # update signature and type annotations of wrapped function
    wrapper.__signature__ = signature(wrapper).replace(
        parameters=[
            p.replace(kind=Parameter.KEYWORD_ONLY)
            for p in (fn_sig.parameters if input_parameters is None else input_parameters).values()
        ],
        return_annotation=fn_sig.return_annotation if return_annotation is None else return_annotation,
    )
    wrapper.__annotations__.update({k: v.annotation for k, v in input_parameters.items()})
    wrapper.__annotations__["return"] = return_annotation

    wrapper.__unionml_object__ = unionml_obj
    output_task = task(wrapper, task_resolver=task_resolver, **task_kwargs)
    output_task._name = f"{unionml_obj.name}.{wrapper.__name__}"
    return output_task


def is_pytorch_model(model_type: typing.Type):
    return model_type.__bases__[0].__module__.startswith("torch")


def is_keras_model(model_type: typing.Type):
    return model_type.__bases__[0].__module__.startswith("keras")



"""unionml cli."""

import copy
import json
import os
import sys
from enum import Enum
from pathlib import Path

import click
import typer
import uvicorn
from cookiecutter.main import cookiecutter

from unionml.remote import get_app_version, get_model, get_model_execution

sys.path.append(os.curdir)

app = typer.Typer()


IMAGE_PREFIX = "unionml"
FLYTE_SANDBOX_CONTAINER_NAME = "flyte-sandbox"


class AppTemplate(str, Enum):
    basic = "basic"
    basic_aws_lambda = "basic-aws-lambda"


@app.command()
def init(
    app_name: str,
    template: AppTemplate = typer.Option(
        AppTemplate.basic,
        "--template",
        "-t",
        help="template to use for initializing your app.",
    ),
):
    """Initialize a UnionML project."""
    config = {
        "app_name": app_name,
    }
    cookiecutter(
        str(Path(__file__).parent / "templates" / template.value),
        no_input=True,
        extra_context=config,
    )


@app.command()
def deploy(app: str):
    """Deploy model to a Flyte backend."""
    typer.echo(f"[unionml] deploying {app}")
    model = get_model(app)
    model.remote_deploy()


@app.command()
def train(
    app: str,
    inputs: str = typer.Option(None, "--inputs", "-i", help="json string of inputs to pass into training workflow"),
    app_version: str = typer.Option(None, "--app-version", "-v", help="app version"),
):
    """Train a model."""
    typer.echo(f"[unionml] app: {app} - training model")
    model = get_model(app)
    train_inputs = {}
    if inputs:
        train_inputs.update(json.loads(inputs))
    model.remote_train(app_version, **train_inputs)
    assert model.artifact is not None
    typer.echo("[unionml] training completed with model artifacts:")
    typer.echo(f"[unionml] model object: {model.artifact.model_object}")
    typer.echo(f"[unionml] model metrics: {model.artifact.metrics}")


@app.command()
def predict(
    app: str,
    inputs: str = typer.Option(None, "--inputs", "-i", help="json string of inputs tp pass into predict workflow"),
    features: Path = typer.Option(None, "--features", "-f", help="generate predictions for this feature"),
    app_version: str = typer.Option(None, "--app-version", "-v", help="app version"),
    model_version: str = typer.Option(None, "--model-version", "-m", help="model version"),
):
    """Generate prediction."""
    typer.echo(f"[unionml] app: {app} - generating predictions")
    model = get_model(app)

    prediction_inputs = {}
    if inputs:
        prediction_inputs.update(json.loads(inputs))
    elif features:
        prediction_inputs.update({"features": model._dataset.get_features(features)})

    predictions = model.remote_predict(app_version, model_version, wait=True, **prediction_inputs)
    typer.echo(f"[unionml] predictions: {predictions}")


@app.command("list-model-versions")
def list_model_versions(
    app: str,
    app_version: str = typer.Option(None, "--app-version", "-v", help="app version"),
    limit: int = typer.Option(
        10, "--limit", help="Maximum number of model versions to list, sorted in descending order of time of execution."
    ),
):
    """List all model versions."""

    model = get_model(app)
    app_version = app_version or get_app_version()
    typer.echo(f"[unionml] app: {app} - listing model versions for app version={app_version}")
    for model_version in model.remote_list_model_versions(app_version, limit):
        typer.echo(f"- {model_version}")


@app.command("fetch-model")
def fetch_model(
    app: str,
    app_version: str = typer.Option(None, "--app-version", "-v", help="app version"),
    model_version: str = typer.Option("latest", "--model-version", "-m", help="model version"),
    output_file: str = typer.Option(None, "--output-file", "-o", help="output file path"),
    kwargs: str = typer.Option(None, "--kwargs", help="json string of kwargs to pass into model.save"),
):
    """Fetch a model object from the remote backend."""
    model = get_model(app)
    app_version = app_version or get_app_version()
    execution = get_model_execution(model, app_version, model_version=model_version)
    model.remote_load(execution)
    save_kwargs = {}
    if kwargs is not None:
        save_kwargs = json.loads(kwargs)
    model.save(output_file, **save_kwargs)
    typer.echo(f"[unionml] app: {app} - saving model version {execution.id.name} to {output_file}")


@app.callback()
def callback():
    """unionml command-line tool."""


def serve_command():
    """Modify the uvicorn.main entrypoint for unionml app serving."""
    fn = copy.deepcopy(uvicorn.main)
    fn.short_help = "Serve an unionml model."
    fn.help = (
        "Serve an unionml model using uvicorn. This command uses the main uvicorn entrypoint with an additional "
        "--model-path argument.\n\nFor more information see: https://www.uvicorn.org/#command-line-options"
    )

    option = click.Option(param_decls=["--model-path"], default=None, help="model path to use for serving", type=Path)
    fn.params.append(option)

    callback = fn.callback

    def custom_callback(**kwargs):
        if os.getenv("UNIONML_MODEL_PATH"):
            typer.echo(
                f"UNIONML_MODEL_PATH environment variable is set to {os.getenv('UNIONML_MODEL_PATH')}. "
                "Please unset this variable before running `unionml serve`.",
                err=True,
            )
            raise typer.Exit(code=1)

        model_path = kwargs.pop("model_path")
        if model_path is not None:
            if not model_path.exists():
                typer.echo(f"Model path {model_path} not found.", err=True)
                raise typer.Exit(code=1)
            os.environ["UNIONML_MODEL_PATH"] = str(model_path)
        return callback(**kwargs)

    fn.callback = custom_callback
    return fn


# convert typer app to click object to define a "serve" command
# that uses the uvicorn.main entrypoint under the hood:
# https://typer.tiangolo.com/tutorial/using-click/#combine-click-and-typer
app = typer.main.get_command(app)
app.add_command(serve_command(), name="serve")


if __name__ == "__main__":
    app()



from setuptools import find_packages, setup

__version__ = "0.1.0b0"

with open("README.md") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    install_requires = [x.strip() for x in f.readlines()]

setup(
    name="unionml",
    version=__version__,
    author="unionai-oss",
    author_email="info@union.ai",
    description="The easiest way to build and deploy machine learning microservices.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache",
    keywords=["machine-learning", "artificial-intelligence", "microservices"],
    data_files=[("", ["LICENSE"])],
    packages=find_packages(exclude=["tests*"]),
    package_data={
        "unionml": ["py.typed"],
        "templates": ["*"],
    },
    python_requires=">=3.7",
    platforms="any",
    install_requires=install_requires,
    entry_points={"console_scripts": ["unionml = unionml.cli:app"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
)



"""Utilities for the FastAPI integration."""

import os
from typing import Any, Dict, List, Optional, Union

from fastapi import Body, FastAPI, HTTPException
from fastapi.responses import HTMLResponse
from pydantic import BaseModel

from unionml.model import Model, ModelArtifact
from unionml.remote import get_model_artifact


def serving_app(model: Model, app: FastAPI, remote: bool = False, model_version: str = "latest"):
    @app.on_event("startup")
    async def setup_model():
        model_path = os.getenv("UNIONML_MODEL_PATH")
        if model.artifact is None:
            if not remote:
                if model.artifact is None and model_path is None:
                    raise ValueError(
                        "Model artifact path not specified. Make sure to specify the unionml serve --model-path in "
                        "the option when starting the unionml prediction service in local mode."
                    )
                model.artifact = ModelArtifact(model.load(model_path))
            else:
                model.artifact = get_model_artifact(model, model_version=model_version)

    @app.get("/", response_class=HTMLResponse)
    def root():
        return """
            <html>
                <head>
                    <title>unionml</title>
                </head>
                <body>
                    <h1>unionml</h1>
                    <p>The easiest way to build and deploy models</p>
                </body>
            </html>
        """

    @app.post("/predict")
    async def predict(
        inputs: Optional[Union[Dict[str, Any], BaseModel]] = Body(None),
        features: Optional[List[Dict[str, Any]]] = Body(None),
    ):
        if inputs is None and features is None:
            raise HTTPException(status_code=500, detail="inputs or features must be supplied.")

        workflow_inputs: Dict[str, Any] = {}
        if model._dataset.reader_return_type is not None:
            # convert raw features to whatever the output type of the reader is.
            features = model._dataset.get_features(features)
        workflow_inputs.update(inputs if inputs else {"features": features})

        return model.predict(**workflow_inputs)
